@@ -30,7 +30,7 @@
 //! ```
 
 use failure::Error;
-use log::{debug, warn};
+use log::{debug, trace, warn};
 use reqwest::blocking::Client;
 use reqwest::Url;
 
@@ -220,7 +220,7 @@ impl SimpleSymbolSupplier {
 impl SymbolSupplier for SimpleSymbolSupplier {
     fn locate_symbols(&self, module: &dyn Module) -> SymbolResult {
         if let Some(rel_path) = relative_symbol_path(module, "sym") {
-            for ref path in self.paths.iter() {
+            for path in self.paths.iter() {
                 let test_path = path.join(&rel_path);
                 if fs::metadata(&test_path).ok().map_or(false, |m| m.is_file()) {
                     return SymbolFile::from_file(&test_path)
@@ -233,6 +233,33 @@ impl SymbolSupplier for SimpleSymbolSupplier {
     }
 }
 
+/// A SymbolSupplier that maps module names (code_files) to an in-memory string.
+///
+/// Intended for mocking symbol files in tests.
+#[derive(Default, Debug, Clone)]
+pub struct StringSymbolSupplier {
+    modules: HashMap<String, String>,
+}
+
+impl StringSymbolSupplier {
+    /// Make a new StringSymbolSupplier with no modules.
+    pub fn new(modules: HashMap<String, String>) -> Self {
+        Self { modules }
+    }
+}
+
+impl SymbolSupplier for StringSymbolSupplier {
+    fn locate_symbols(&self, module: &dyn Module) -> SymbolResult {
+        if let Some(symbols) = self.modules.get(&*module.code_file()) {
+            return SymbolFile::from_bytes(symbols.as_bytes())
+                .map(SymbolResult::Ok)
+                .unwrap_or_else(SymbolResult::LoadError);
+        }
+
+        SymbolResult::NotFound
+    }
+}
+
 /// An implementation of `SymbolSupplier` that loads Breakpad text-format symbols from HTTP
 /// URLs.
 ///
@@ -300,7 +327,7 @@ fn fetch_symbol_file(
     rel_path: &str,
     cache: &Path,
 ) -> Result<Vec<u8>, Error> {
-    let url = base_url.join(&rel_path)?;
+    let url = base_url.join(rel_path)?;
     debug!("Trying {}", url);
     let mut res = client.get(url).send()?.error_for_status()?;
     let mut buf = vec![];
@@ -320,7 +347,7 @@ impl SymbolSupplier for HttpSymbolSupplier {
             res @ SymbolResult::Ok(_) | res @ SymbolResult::LoadError(_) => res,
             SymbolResult::NotFound => {
                 if let Some(rel_path) = relative_symbol_path(module, "sym") {
-                    for ref url in self.urls.iter() {
+                    for url in self.urls.iter() {
                         if let Ok(buf) =
                             fetch_symbol_file(&self.client, url, &rel_path, &self.cache)
                         {
@@ -360,6 +387,8 @@ pub trait FrameWalker {
     fn get_callee_register(&self, name: &str) -> Option<u64>;
     /// Set the value of a register for the caller's frame.
     fn set_caller_register(&mut self, name: &str, val: u64) -> Option<()>;
+    /// Explicitly mark one of the caller's registers as invalid.
+    fn clear_caller_register(&mut self, name: &str);
     /// Set whatever registers in the caller should be set based on the cfa (e.g. rsp).
     fn set_cfa(&mut self, val: u64) -> Option<()>;
     /// Set whatever registers in the caller should be set based on the return address (e.g. rip).
@@ -479,7 +508,7 @@ impl Symbolizer {
     ) -> Option<String> {
         let k = (debug_file, debug_id);
         let mut frame = SimpleFrame::with_instruction(address);
-        self.fill_symbol(&k, &mut frame);
+        self.fill_symbol(&k, &mut frame).ok()?;
         frame.function
     }
 
@@ -488,6 +517,9 @@ impl Symbolizer {
     /// using a minidump module, you can use [`SimpleModule`][simplemodule] and
     /// [`SimpleFrame`][simpleframe].
     ///
+    /// An Error indicates that no symbols could be found for the relevant
+    /// module.
+    ///
     /// # Examples
     ///
     /// ```
@@ -499,7 +531,7 @@ impl Symbolizer {
     /// let symbolizer = Symbolizer::new(supplier);
     /// let m = SimpleModule::new("test_app.pdb", "5A9832E5287241C1838ED98914E9B7FF1");
     /// let mut f = SimpleFrame::with_instruction(0x1010);
-    /// symbolizer.fill_symbol(&m, &mut f);
+    /// let _ = symbolizer.fill_symbol(&m, &mut f);
     /// assert_eq!(f.function.unwrap(), "vswprintf");
     /// assert_eq!(f.source_file.unwrap(), r"c:\program files\microsoft visual studio 8\vc\include\swprintf.inl");
     /// assert_eq!(f.source_line.unwrap(), 51);
@@ -507,11 +539,21 @@ impl Symbolizer {
     ///
     /// [simplemodule]: struct.SimpleModule.html
     /// [simpleframe]: struct.SimpleFrame.html
-    pub fn fill_symbol(&self, module: &dyn Module, frame: &mut dyn FrameSymbolizer) {
+    #[allow(clippy::result_unit_err)]
+    pub fn fill_symbol(
+        &self,
+        module: &dyn Module,
+        frame: &mut dyn FrameSymbolizer,
+    ) -> Result<(), ()> {
         let k = key(module);
         self.ensure_module(module, &k);
         if let Some(SymbolResult::Ok(ref sym)) = self.symbols.borrow().get(&k) {
-            sym.fill_symbol(module, frame)
+            sym.fill_symbol(module, frame);
+            Ok(())
+        } else {
+            // Treat it as an error if we can't find a symbol file (this allows us to distinguish
+            // between "we had no symbols" and "we had symbols and this frame just had no matches").
+            Err(())
         }
     }
 
@@ -519,14 +561,16 @@ impl Symbolizer {
         let k = key(module);
         self.ensure_module(module, &k);
         if let Some(SymbolResult::Ok(ref sym)) = self.symbols.borrow().get(&k) {
+            trace!("unwind: found symbols for address, searching for cfi entries");
             sym.walk_frame(module, walker)
         } else {
+            trace!("unwind: couldn't find symbols for address, cannot use cfi");
             None
         }
     }
 
     fn ensure_module(&self, module: &dyn Module, k: &ModuleKey) {
-        if !self.symbols.borrow().contains_key(&k) {
+        if !self.symbols.borrow().contains_key(k) {
             let res = self.supplier.locate_symbols(module);
             debug!("locate_symbols for {}: {}", module.code_file(), res);
             self.symbols.borrow_mut().insert(k.clone(), res);
@@ -728,7 +772,7 @@ FUNC 1000 30 10 some func
 ",
         );
         let mut f1 = SimpleFrame::with_instruction(0x1010);
-        symbolizer.fill_symbol(&m1, &mut f1);
+        symbolizer.fill_symbol(&m1, &mut f1).unwrap();
         assert_eq!(f1.function.unwrap(), "some func");
         assert_eq!(f1.function_base.unwrap(), 0x1000);
         assert_eq!(f1.source_file.unwrap(), "foo.c");
@@ -745,7 +789,7 @@ FUNC 1000 30 10 some func
         let m2 = SimpleModule::new("bar.pdb", "ffff0000");
         let mut f2 = SimpleFrame::with_instruction(0x1010);
         // No symbols present, should not find anything.
-        symbolizer.fill_symbol(&m2, &mut f2);
+        assert!(symbolizer.fill_symbol(&m2, &mut f2).is_err());
         assert!(f2.function.is_none());
         assert!(f2.function_base.is_none());
         assert!(f2.source_file.is_none());
@@ -759,7 +803,7 @@ FUNC 1000 30 10 another func
 1000 30 7 53
 ",
         );
-        symbolizer.fill_symbol(&m2, &mut f2);
+        assert!(symbolizer.fill_symbol(&m2, &mut f2).is_err());
         assert!(f2.function.is_none());
         assert!(f2.function_base.is_none());
         assert!(f2.source_file.is_none());
